/* Parse command line, set up command arguments Unix-style, and call function.
 * Note: argument is modified (delimiters are overwritten with nulls)
 *
 * Copyright 1991 Phil Karn, KA9Q
 *
 * Improved error handling by Brian Boesch of Stanford University
 * Feb '91 - Bill Simpson
 *              bit16cmd for PPP
 * Mar '91 - Glenn McGregor
 *              handle string escaped sequences
 */
#include <common.h>
#include "global.h"
#include "cmdparse.h"
  
struct boolcmd {
    char *str;      /* Token */
    int val;        /* Value */
};
  
static struct boolcmd Boolcmds[] = {
    {"y",            1},      /* Synonyms for "true" */
    {"yes",          1},
    {"true",         1},
    {"on",           1},
    {"1",            1},
    {"set",          1},
    {"enable",       1},
  
    {"n",            0},      /* Synonyms for "false" */
    {"no",           0},
    {"false",        0},
    {"off",          0},
    {"0",            0},
    {"clear",        0},
    {"disable",      0},
    {NULLCHAR,		0}
};
  
static char *stringparse __ARGS((char *line));
  
static char *
stringparse(line)
char *line;
{
    register char *cp = line;
    unsigned long num;
  
    while ( *line != '\0' && *line != '\"' ) {
        if ( *line == '\\' ) {
            line++;
            switch ( *line++ ) {
                case 'n':
                    *cp++ = '\n';
                    break;
                case 't':
                    *cp++ = '\t';
                    break;
                case 'v':
                    *cp++ = '\v';
                    break;
                case 'b':
                    *cp++ = '\b';
                    break;
                case 'r':
                    *cp++ = '\r';
                    break;
                case 'f':
                    *cp++ = '\f';
                    break;
                case 'a':
                    *cp++ = '\a';
                    break;
                case '\\':
                    *cp++ = '\\';
                    break;
                case '\?':
                    *cp++ = '\?';
                    break;
                case '\'':
                    *cp++ = '\'';
                    break;
                case '\"':
                    *cp++ = '\"';
                    break;
                case 'x':
		    line--;
                    num = strtoul(line, &line, 16 );
                    *cp++ = (char) num;
                    break;
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
		    line--;
                    num = strtoul(line, &line, 8 );
                    *cp++ = (char) num;
                    break;
                case '\0':
                    return NULLCHAR;
                default:
                    *cp++ = *(line - 1);
                    break;
            };
        } else {
            *cp++ = *line++;
        }
    }
  
    if ( *line == '\"' )
        line++;         /* skip final quote */
    *cp = '\0';             /* terminate string */
    return line;
}
  
int
cmdparse(cmds,line,p)
struct cmds cmds[];
register char *line;
void *p;
{
    struct cmds *cmdp;
    char *argv[NARG],*cp;
    int argc,i;
  
    /* Remove cr/lf */
    rip(line);
  
    for(argc = 0;argc < NARG;argc++)
        argv[argc] = NULLCHAR;
  
    for(argc = 0;argc < NARG;){
        register int qflag = FALSE;
  
        /* Skip leading white space */
        while(*line == ' ' || *line == '\t')
            line++;
        if(*line == '\0')
            break;
        /* return if comment character first non-white */
        if ( argc == 0  &&  *line == '#' )
            return 0;
        /* Check for quoted token */
        if(*line == '"'){
            line++; /* Suppress quote */
            qflag = TRUE;
        }
        argv[argc++] = line;    /* Beginning of token */
  
        if(qflag){
            /* Find terminating delimiter */
            if((line = stringparse(line)) == NULLCHAR){
                return -1;
            }
        } else {
            /* Find space or tab. If not present,
             * then we've already found the last
             * token.
             */
            for(cp=line;*cp;cp++) {
                if(*cp == ' ' || *cp == '\t')
                    break;
            }
        /* if we didn't reach end of line yet, cutoff this arg here */
            if(*cp == '\0')
                break;
            *cp++ = '\0';
            line = cp;
        }
    }
    if (argc < 1) {         /* empty command line */
        argc = 1;
        argv[0] = "";
    }
  
    /* Look up command in table; prefix matches are OK */
    /* Not case sensitive anymore - WG7J */
    for(cmdp = cmds;cmdp->name != NULLCHAR;cmdp++){
        if(strncasecmp(argv[0],cmdp->name,strlen(argv[0])) == 0)
            break;
    }
    if(cmdp->name == NULLCHAR) {
        if(cmdp->argc_errmsg != NULLCHAR)
            printf("%s\n",cmdp->argc_errmsg);
        return -1;
    } else {
        if(argc < cmdp->argcmin) {
            /* Insufficient arguments */
            printf("Usage: %s\n",cmdp->argc_errmsg);
            i = -1;  /* return code */
        } else {
	    i=(*cmdp->func)(argc,argv,p);
        }
    }
    return i;  /* 0 => OK, else err */
}
  
/* Call a subcommand based on the first token in an already-parsed line */
int
subcmd(tab,argc,argv,p)
struct cmds tab[];
int argc;
char *argv[];
void *p;
{
    register struct cmds *cmdp;
    int found = 0;
    int i;
  
    /* Strip off first token and pass rest of line to subcommand */
    if (argc < 2) {
        if (argc < 1)
            puts("SUBCMD - Don't know what to do?\n");
        else
            printf("\"%s\" - takes at least one argument\n",argv[0]);
        return -1;
    }
    argc--;
    argv++;
    for(cmdp = tab;cmdp->name != NULLCHAR;cmdp++){
    /* Not case sensitive anymore - WG7J */
        if(strncasecmp(argv[0],cmdp->name,strlen(argv[0])) == 0){
            found = 1;
            break;
        }
    }
    if(!found){
        char buf[77];
        memset(buf,' ',sizeof(buf));
        buf[75] = '\n';
        buf[76] = '\0';
        puts("valid subcommands:\n");
        for(i = 0, cmdp = tab;cmdp->name != NULLCHAR; cmdp++, i = (i+1)%5) {
            strncpy(&buf[i*15],cmdp->name,strlen(cmdp->name));
            if(i == 4) {
                puts(buf);
                memset(buf,' ',sizeof(buf));
                buf[75] = '\n';
                buf[76] = '\0';
            }
        }
        if(i != 0)
            puts(buf);
        return -1;
    }
    if(argc < cmdp->argcmin){
        if(cmdp->argc_errmsg != NULLCHAR)
            printf("Usage: %s\n",cmdp->argc_errmsg);
        return -1;
    }
    return (*cmdp->func)(argc,argv,p);
}
  
/* Subroutine for setting and displaying boolean flags */
int
setbool(var,label,argc,argv)
int *var;
char *label;
int argc;
char *argv[];
{
    struct boolcmd *bc;
  
    if(argc < 2){
        printf("%s: %s\n",label,*var ? "on":"off");
        return 1;
    }
    for(bc = Boolcmds;bc->str != NULLCHAR;bc++){
        if(strcasecmp(argv[1],bc->str) == 0){
            *var = bc->val;
            return 0;
        }
    }
    puts("Valid options:");
    for(bc = Boolcmds;bc->str != NULLCHAR;bc++)
        printf(" %s",bc->str);
    putc('\n');
    return 1;
}
  
  
/* Subroutine for setting and displaying long variables */
int
setlong(var,label,argc,argv)
long *var;
char *label;
int argc;
char *argv[];
{
    if(argc < 2){
        printf("%s: %ld\n",label,*var);
        return 1;
    } else {
        *var = atol(argv[1]);
        return 0;
    }
  
}
/* Subroutine for setting and displaying short variables */
int
setshort(var,label,argc,argv)
unsigned short *var;
char *label;
int argc;
char *argv[];
{
    if(argc < 2){
        printf("%s: %u\n",label,*var);
        return 1;
    } else {
        *var = atoi(argv[1]);
        return 0;
    }
}

/* Subroutine for setting and displaying short HEX variables */
int
setshorthex(var,label,argc,argv)
unsigned short *var;
char *label;
int argc;
char *argv[];
{
    if(argc < 2){
        printf("%s: %x\n",label,*var);
        return 1;
    } else {
        *var = htoi(argv[1]);
        return 0;
    }
}

/* Subroutine for setting and displaying integer variables */
int
setint(var,label,argc,argv)
int *var;
char *label;
int argc;
char *argv[];
{
    if(argc < 2){
        printf("%s: %i\n",label,*var);
        return 1;
    } else {
        *var = atoi(argv[1]);
        return 0;
    }
  
}
  
/* Subroutine for setting and displaying unsigned integer variables */
int
setuns(var,label,argc,argv)
unsigned *var;
char *label;
int argc;
char *argv[];
{
    if(argc < 2){
        printf("%s: %u\n",label,*var);
        return 1;
    } else {
        *var = atoi(argv[1]);
        return 0;
    }
}
  
/* Subroutine for setting and displaying int variables (with range check) */
int
setintrc(var, label, argc, argv, minval, maxval)
int16 *var;
char *label;
int argc;
char *argv[];
int minval;
int16 maxval;
{
    int tmp;
  
    if (argc < 2)
        printf("%s: %u\n", label, *var);
    else {
        tmp = atoi(argv[1]);
        if (isalpha(*argv[1]) || tmp < minval || tmp > maxval) {
            printf("%s must be %i..%i\n", label, minval, maxval);
            return 1;
        }
        *var = (int16)tmp;
    }
    return 0;
}
  
/* replace terminating end of line marker(s) with null */
void
rip(s)
register char *s;
{
    register char *cp;
  
    if((cp = strpbrk(s,"\r\n")) != NULLCHAR)	/* n5knx: was "\n" */
        *cp = '\0';
}
  
